from typing import AsyncGenerator

import entities

from .base import GetBase, ListBase


class Episode(ListBase[entities.Episode], GetBase[entities.Episode]):
    entity_class = entities.Episode
    endpoint = "/v1/episode"
    list_body = {"order": "desc", "limit": 20}
    get_params_key = "eid"

    def list_by_podcast(
        self, podcast_id: str
    ) -> AsyncGenerator[entities.Episode, None]:
        return self.list({"pid": podcast_id})



from typing import AsyncGenerator

import entities

from .base import ListBase


class PodcastSearch(ListBase[entities.Podcast]):
    entity_class = entities.Podcast
    endpoint = "/v1/search"
    list_suffix = "create"
    list_body = {"limit": 20}

    def search(self, keyword: str) -> AsyncGenerator[entities.Podcast, None]:
        return self.list({"keyword": keyword, "type": "PODCAST"})


class EpisodeSearch(ListBase[entities.Episode]):
    entity_class = entities.Episode
    endpoint = "/v1/search"
    list_suffix = "create"
    list_body = {"limit": 20}

    def search(self, keyword: str) -> AsyncGenerator[entities.Episode, None]:
        return self.list({"keyword": keyword, "type": "EPISODE"})


class UserSearch(ListBase[entities.User]):
    entity_class = entities.User
    endpoint = "/v1/search"
    list_suffix = "create"
    list_body = {"limit": 20}

    def search(self, keyword: str) -> AsyncGenerator[entities.User, None]:
        return self.list({"keyword": keyword, "type": "USER"})



import entities

from .base import ListBase


class FavoritedEpisode(ListBase[entities.Episode]):
    entity_class = entities.Episode
    endpoint = "/v1/favorite"


class FavoritedComment(ListBase[entities.Comment]):
    entity_class = entities.Comment
    endpoint = "/v1/comment/collect"



from datetime import datetime
from typing import Any

from .base import Base
from .podcast import Podcast


class Episode(Base):
    def __init__(
        self,
        eid: str,
        pid: str,
        title: str,
        description: str = "",
        shownotes: str = "",
        duration: int = 0,
        image: dict[str, Any] = {},
        enclosure: dict[str, str] = {},
        mediaKey: str = "",
        media: dict[str, Any] = {},
        playCount: int = 0,
        playTime: int = 0,
        clapCount: int = 0,
        commentCount: int = 0,
        favoriteCount: int = 0,
        isPlayed: bool = False,
        isFinished: bool = False,
        isPicked: bool = False,
        isFavorited: bool = False,
        pubDate: str = "",
        podcast: dict = {},
        **kwargs,
    ):
        super().__init__(id=eid)
        self.pid = pid
        self.title = title
        self.description = description
        self.shownotes = shownotes
        self.duration = duration
        self.image = image
        self.enclosure = enclosure
        self.media_key = mediaKey
        self.media = media
        self.play_count = playCount
        self.play_time = playTime
        self.clap_count = clapCount
        self.comment_count = commentCount
        self.favorite_count = favoriteCount
        self.is_played = isPlayed
        self.is_finished = isFinished
        self.is_picked = isPicked
        self.is_favorited = isFavorited
        self.pub_date = datetime.fromisoformat(pubDate) if pubDate else None
        self.podcast = Podcast(**podcast) if podcast else None
        # set rest keys as attributes
        for key, value in kwargs.items():
            setattr(self, key, value)

    def __repr__(self) -> str:
        return f"Episode(eid={self.id}, pid={self.pid}, title={self.title})"



from .base import ListBase


class Playlist(ListBase[str]):
    entity_class = str
    endpoint = "/v1/playlist"
    list_suffix = "pull"

    def _extract_data(self, json_resp: dict) -> list[dict]:
        data = json_resp.get("data", {})
        return data.get("list", [])



from typing import AsyncGenerator, Literal

import entities

from .base import ListBase


class Comment(ListBase[entities.Comment]):
    entity_class = entities.Comment
    endpoint = "/v1/comment"
    list_body = {"limit": 20}
    list_suffix = "list-primary"

    def list_by_episode(
        self, episode_id: str, order: Literal["HOT", "TIMESTAMP", "TIME"] = "HOT"
    ) -> AsyncGenerator[entities.Comment, None]:
        return self.list(
            {"order": order, "owner": {"type": "EPISODE", "id": episode_id}}
        )


class Reply(ListBase[entities.Comment]):
    entity_class = entities.Comment
    endpoint = "/v1/comment"
    list_body = {}
    list_suffix = "list-thread"

    def list_by_comment(
        self, comment_id: str, order: Literal["SMART", "TIME"] = "SMART"
    ) -> AsyncGenerator[entities.Comment, None]:
        return self.list({"primaryCommentId": comment_id, "order": order})



from datetime import datetime
from typing import Any, Literal

from .base import Base
from .user import User


class Podcast(Base):
    def __init__(
        self,
        pid: str,
        title: str,
        author: str = "",
        description: str = "",
        brief: str = "",
        image: dict[str, str] = {},
        color: dict[str, str] = {},
        subscriptionCount: int = 0,
        episodeCount: int = 0,
        latestEpisodePubDate: str = "",
        podcasters: list[dict[str, Any]] = [],
        subscritionStatus: Literal["ON", "OFF"] = "OFF",
        **kwargs,
    ):
        super().__init__(id=pid)
        self.title = title
        self.author = author
        self.description = description
        self.brief = brief
        self.image = image
        self.color = color
        self.subscription_count = subscriptionCount
        self.episode_count = episodeCount
        self.latest_episode_pub_date = (
            datetime.fromisoformat(latestEpisodePubDate)
            if latestEpisodePubDate
            else None
        )
        self.podcasters = [
            User(**podcaster) if isinstance(podcaster, dict) else podcaster
            for podcaster in podcasters
        ]
        self.subscription_status = subscritionStatus
        # set rest keys as attributes
        for key, value in kwargs.items():
            setattr(self, key, value)

    def __repr__(self) -> str:
        return f"Podcast(pid={self.id}, title={self.title})"



import asyncio

import httpx
import qrcode
from managers import Token


class Login:
    def __init__(self) -> None:
        self.client = httpx.AsyncClient(
            base_url="https://podcaster-api.xiaoyuzhoufm.com/v1/auth"
        )
        self.id = None

    async def _get_login_url(self) -> str:
        resp = await self.client.post(
            "/qrcode/create",
            json={"clientId": "podcaster-platform"},
            headers={"X-Midway-App-Id": "v6worU4NnWyL"},
        )
        data = resp.raise_for_status().json()
        self.id, url = data.values()
        return url

    async def _check_login(self) -> tuple[str, str] | None:
        resp = await self.client.post("/qrcode/login", json={"id": self.id})
        data = resp.raise_for_status().json()
        if data["status"] != "USED":
            return None

        access_token = resp.headers.get(Token.access_key)
        refresh_token = resp.headers.get(Token.refresh_key)
        assert access_token and refresh_token, "Login failed, tokens not found."
        return (access_token, refresh_token)

    async def login(self) -> tuple[str, str]:
        url = await self._get_login_url()
        qr = qrcode.QRCode()
        qr.add_data(url)
        qr.print_tty()

        headers = await self._check_login()
        while headers is None:
            await asyncio.sleep(1)
            headers = await self._check_login()

        await self.client.aclose()
        return headers



import asyncio
import json
import os
import time

import httpx


class Token:
    access_key = "X-Jike-Access-Token"
    refresh_key = "X-Jike-Refresh-Token"


class TokenManager:
    def __init__(self):
        self.token_path = os.path.join(
            os.path.expanduser("~"), ".local", "state", "cosmos-wormhole", "token.json"
        )

    def get_token(self) -> dict:
        if not os.path.exists(self.token_path):
            return {}

        with open(self.token_path, "r", encoding="utf-8") as f:
            tokens = json.load(f)
        expires_at = tokens.get("expiresAt")
        if expires_at and expires_at < time.time():
            return {}

        access_token = tokens.get("accessToken")
        refresh_token = tokens.get("refreshToken")
        assert access_token and refresh_token, "Tokens not found."
        return {
            Token.access_key: access_token,
            Token.refresh_key: refresh_token,
        }

    def save_token(self, access_token: str, refresh_token: str) -> None:
        tokens = {
            "accessToken": access_token,
            "refreshToken": refresh_token,
            "expiresAt": int(time.time()) + 60 * 60 * 24 * 30,  # 30 days
        }
        os.makedirs(os.path.dirname(self.token_path), exist_ok=True)
        with open(self.token_path, "w", encoding="utf-8") as f:
            json.dump(tokens, f, ensure_ascii=False, indent=4)

    def delete_token(self) -> None:
        if os.path.exists(self.token_path):
            os.remove(self.token_path)

    async def refresh_token(self, client: httpx.AsyncClient) -> tuple[str, str]:
        resp = await client.post("/app_auth_tokens.refresh")
        access_token = resp.headers.get(Token.access_key)
        refresh_token = resp.headers.get(Token.refresh_key)
        return access_token, refresh_token

    async def periodic_refresh_token(
        self, client: httpx.AsyncClient, interval: int, queue: asyncio.Queue
    ) -> None:
        if not os.path.exists(self.token_path):
            return
        try:
            while True:
                await asyncio.sleep(interval)
                access_token, refresh_token = await self.refresh_token(client)
                if access_token and refresh_token:
                    await queue.put((access_token, refresh_token))
        except asyncio.CancelledError:
            pass



from datetime import datetime

from .base import Base
from .user import User


class Comment(Base):
    def __init__(
        self,
        id: str,
        owner: dict[str, str] = {},
        pid: str = "",
        author: dict = {},
        text: str = "",
        level: int = 0,
        likeCount: int = 0,
        createdAt: str = "",
        entities: list[dict] = [],
        replyCount: int = 0,
        threadReplyCount: int = 0,
        replies: list[dict] = [],
        replyToComment: dict = {},
        # HELP WANTED: voice comment structure, welcome to contribute
        **kwargs,
    ):
        super().__init__(id=id)
        self.owner = owner
        self.pid = pid
        self.author = User(**author) if author else None
        self.text = text
        self.level = level
        self.like_count = likeCount
        self.created_at = datetime.fromisoformat(createdAt) if createdAt else None
        self.entities = entities
        self.reply_count = replyCount
        self.thread_reply_count = threadReplyCount
        self.replies = [Comment(**reply) for reply in replies]
        self.reply_to_comment = Comment(**replyToComment) if replyToComment else None
        # set rest keys as attributes
        for key, value in kwargs.items():
            setattr(self, key, value)

    def __repr__(self) -> str:
        return f"Comment(id={self.id}, text={self.text})"



from .comment import Comment, Reply
from .episode import Episode
from .favorite import FavoritedComment, FavoritedEpisode
from .history import History
from .inbox import Inbox
from .login import Login
from .playlist import Playlist
from .podcast import Podcast
from .profile import Profile
from .search import EpisodeSearch, PodcastSearch, UserSearch
from .subscription import Subscription
from .user_relation import Followee, Follower

__all__ = [
    "Login",
    "Subscription",
    "Episode",
    "Comment",
    "Reply",
    "Playlist",
    "Podcast",
    "Inbox",
    "Profile",
    "Followee",
    "Follower",
    "History",
    "FavoritedEpisode",
    "FavoritedComment",
    "PodcastSearch",
    "EpisodeSearch",
    "UserSearch",
]



import entities

from .base import ListBase


class Subscription(ListBase[entities.Podcast]):
    entity_class = entities.Podcast
    endpoint = "/v1/subscription"
    list_body = {"sortBy": "subscribedAt", "sortOrder": "desc", "limit": 20}



import asyncio

import httpx
from __about__ import __version__
from endpoints import *
from managers import *


class Client:
    user_agent = f"CosmosWormhole/{__version__}"
    device_id = "cosmos-wormhole"

    def __init__(self, base_url: str = "https://api.xiaoyuzhoufm.com") -> None:
        self.client = httpx.AsyncClient(
            base_url=base_url,
            headers={"user-agent": self.user_agent, "x-jike-device-id": self.device_id},
        )
        self.token_manager = TokenManager()
        self._token_update_queue = asyncio.Queue(1)
        self._token_update_producer: asyncio.Task
        self._token_update_consumer: asyncio.Task

        self.subscription: Subscription
        self.episode: Episode
        self.comment: Comment
        self.reply: Reply
        self.playlist: Playlist
        self.podcast: Podcast
        self.inbox: Inbox
        self.profile: Profile
        self.followee: Followee
        self.follower: Follower
        self.history: History
        self.favorited_episode: FavoritedEpisode
        self.favorited_comment: FavoritedComment
        self.podcast_search: PodcastSearch
        self.episode_search: EpisodeSearch
        self.user_search: UserSearch

    def _init_endpoints(self) -> None:
        self.subscription = Subscription(self.client)
        self.episode = Episode(self.client)
        self.comment = Comment(self.client)
        self.reply = Reply(self.client)
        self.playlist = Playlist(self.client)
        self.podcast = Podcast(self.client)
        self.inbox = Inbox(self.client)
        self.profile = Profile(self.client)
        self.followee = Followee(self.client)
        self.follower = Follower(self.client)
        self.history = History(self.client)
        self.favorited_episode = FavoritedEpisode(self.client)
        self.favorited_comment = FavoritedComment(self.client)
        self.podcast_search = PodcastSearch(self.client)
        self.episode_search = EpisodeSearch(self.client)
        self.user_search = UserSearch(self.client)

    async def close(self) -> None:
        await self._token_update_queue.join()

        tasks = [self.token_update_producer, self.token_update_consumer]
        for task in tasks:
            if task and not task.done():
                task.cancel()

        await asyncio.gather(*tasks, return_exceptions=True)

        await self.client.aclose()

    def _update_token(self, access_token: str, refresh_token: str) -> None:
        if not access_token or not refresh_token:
            raise ValueError("Access token and refresh token cannot be empty.")
        self.token_manager.save_token(
            access_token=access_token, refresh_token=refresh_token
        )
        self.client.headers.update(
            {Token.access_key: access_token, Token.refresh_key: refresh_token}
        )

    async def login(self) -> None:
        headers = self.token_manager.get_token()

        if (
            not headers
            or not headers.get(Token.access_key)
            or not headers.get(Token.refresh_key)
        ):  # first time or no token stored or token invalid, try to login
            access_token, refresh_token = await Login().login()
            self._update_token(access_token, refresh_token)
        else:  # refresh tokens in case of access token expired
            self.client.headers.update(headers)
            access_token, refresh_token = await self.token_manager.refresh_token(
                self.client
            )
            self._update_token(access_token, refresh_token)

        self._after_login()

    def _after_login(self) -> None:
        self.token_update_producer = asyncio.create_task(
            self.token_manager.periodic_refresh_token(
                self.client, 60 * 20, self._token_update_queue
            )  # every 20 minutes
        )
        self.token_update_consumer = asyncio.create_task(self._listen_on_token_update())
        self._init_endpoints()

    async def _listen_on_token_update(self) -> None:
        try:
            while True:
                access_token, refresh_token = await self._token_update_queue.get()
                self._update_token(access_token, refresh_token)
                self._token_update_queue.task_done()
        except asyncio.CancelledError:
            pass


async def main():
    c = Client()
    await c.login()

    print("1. Subscription:")
    pid = None
    eid = None
    podcast_uid = None
    async for podcast in c.subscription.list():
        print(podcast)
        if podcast.title == "史蒂夫说":
            pid = podcast.id
            podcast_uid = podcast.podcasters[0].id if podcast.podcasters else None
            async for episode in c.episode.list_by_podcast(podcast.id):
                print(f"\t{episode}")
                eid = episode.id
                async for comment in c.comment.list_by_episode(episode.id, "HOT"):
                    print(f"\t\t{comment}")
                    async for reply in c.reply.list_by_comment(comment.id, "SMART"):
                        print(f"\t\t\t{reply}")
                    break
                break
            break

    print("2. Podcaast/Episode get:")
    if pid:
        podcast = await c.podcast.get(pid)
        print(podcast)
    if eid:
        episode = await c.episode.get(eid)
        print(episode)

    print("3. Inbox:")
    cnt = 0
    async for episode in c.inbox.list():
        print(episode)
        cnt += 1
        if cnt >= 5:
            break

    print("4. Playlist:")
    cnt = 0
    async for playlist in c.playlist.list():
        print(playlist)
        cnt += 1
        if cnt >= 5:
            break

    print("5. Profile:")
    my_profile = await c.profile.get()
    print(my_profile)

    if podcast_uid:
        print("5.1 Podcaster Profile:")
        podcaster = await c.profile.get(podcast_uid)
        print(podcaster)

    print("6. Followee:")
    cnt = 0
    async for user in c.followee.list_following(my_profile.id):
        print(user)
        cnt += 1
        if cnt >= 5:
            break

    print("7. Follower:")
    cnt = 0
    async for user in c.follower.list_follower(my_profile.id):
        print(user)
        cnt += 1
        if cnt >= 5:
            break

    print("8. History:")
    cnt = 0
    async for episode in c.history.list():
        print(episode)
        cnt += 1
        if cnt >= 5:
            break

    print("9. Favorited Episode:")
    cnt = 0
    async for episode in c.favorited_episode.list():
        print(episode)
        cnt += 1
        if cnt >= 5:
            break

    print("10. Favorited Comment:")
    cnt = 0
    async for comment in c.favorited_comment.list():
        print(comment)
        cnt += 1
        if cnt >= 5:
            break

    print("11. Podcast Search:")
    cnt = 0
    async for podcast in c.podcast_search.search("即刻"):
        print(podcast)
        cnt += 1
        if cnt >= 5:
            break

    print("12. Episode Search:")
    cnt = 0
    async for episode in c.episode_search.search("小宇宙"):
        print(episode)
        cnt += 1
        if cnt >= 5:
            break

    print("13. User Search:")
    cnt = 0
    async for user in c.user_search.search("播客"):
        print(user)
        cnt += 1
        if cnt >= 5:
            break

    await c.close()


if __name__ == "__main__":
    asyncio.run(main())
